//
// C Implementation: eel
//
// Description: 
//
//
// Author: Ryan Cunningham <rcunning@cs.ucla.edu>, (C) 2004
//
// Copyright: Ryan Cunningham
//
//

// a mess of headers here, likely can remove some
#include <sys/types.h>
#include <tree.h>
#include <queue.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <err.h>
#include <assert.h>

#include <eel.h>
#include <eel_util.h>

// eel single instance library data structure
// initialized in eel_initialize()
struct eel_lib_struct eel;

//
// Support functions
//

// --------------------------------------------------------
// free_event - free the event structure
// --------------------------------------------------------
static void free_event(struct eel_event *event)
{
  assert(event);
  free((void*)event);
}

// --------------------------------------------------------
// free_group - free the group structure
// --------------------------------------------------------
static void free_group(struct eel_group *group)
{
  assert(group);
  free((void*)group);
}

// --------------------------------------------------------
// add_to_queue - add event to eel.queue, return -1 on 
// error 0 on success
// --------------------------------------------------------
static int add_to_queue(struct eel_event *event)
{
  if (event && event->state == EEL_CREATED) {
    if (event->type == EEL_TIMER_EVENT) {
      RB_INSERT(event_rb_tree, &eel.timer_rb, event);
    } else {
      TAILQ_INSERT_TAIL(&eel.queue, event, queue_entry);
    }
    // we are waiting on this event now
    event->state = EEL_WAITING;
    return 0;
  }
  return -1;
}

// --------------------------------------------------------
// add_to_fire - add event to eel.fire_queue, return -1 on
// error 0 on success
// --------------------------------------------------------
static int add_to_fire(struct eel_event *event)
{
  if (event && event->state == EEL_WAITING) {
    TAILQ_INSERT_TAIL(&eel.fire_queue, event, fire_queue_entry);
    // this event is queued for firing
    event->state = EEL_FIRE;
    return 0;
  }
  return -1;
}

// --------------------------------------------------------
// add_to_group - 
// --------------------------------------------------------
static int add_to_group(struct eel_group *group, struct eel_event *event)
{
  if (event && group) {
    TAILQ_INSERT_TAIL(&group->event_queue, event, group_queue_entry);
    return 0;
  }
  return -1;
}

// --------------------------------------------------------
// just_fire - call the callback for event
// --------------------------------------------------------
static void just_fire(struct eel_event *event)
{
  // ignore return value from callback function
  (void)event->callback(event->group, event->callback_arg, event->fd);
}

// --------------------------------------------------------
// just_remove - helper, remove event from any eel
// queues (eel.queue, eel.fire_queue, or eel.timer_rb)
// returns 1 if the event was removed, 0 if not
// --------------------------------------------------------
static int just_remove(struct eel_event *event)
{
  if (event && (event->state == EEL_WAITING || event->state == EEL_FIRE)) {
    if (event->type == EEL_TIMER_EVENT) {
      RB_REMOVE(event_rb_tree, &eel.timer_rb, event);
    } else {
      TAILQ_REMOVE(&eel.queue, event, queue_entry);
      // notify the event mechnism that this event is gone
      eel.ops->remove(&eel, event);
    }
    if (event->state == EEL_FIRE) {
      TAILQ_REMOVE(&eel.fire_queue, event, fire_queue_entry);
    }
    // remove the event from its group
    TAILQ_REMOVE(&event->group->event_queue, event, group_queue_entry);
    // this event is now expired because it is not in any queue
    event->state = EEL_EXPIRED;
    return 1;
  }
  return 0;
}

// --------------------------------------------------------
// remove_and_free_ORs - helper, remove all the OR events 
// in the event's OR ring and free them
// --------------------------------------------------------
static void remove_and_free_ORs(struct eel_event *event)
{
  if (event) {
    // remove and free all its ORs but do not free it
    struct eel_event *ring_item;
    RING_FORALLOTHERS(ring_item, event, OR_ring) {
      if (just_remove(ring_item)) {
        free_event(ring_item);
      } else {
        assert(0);
      }
    }
  }
}

// --------------------------------------------------------
// remove_fire_and_free - clean up all event's ORs, fire
// event, then free event
// --------------------------------------------------------
static void remove_fire_and_free(struct eel_event *event)
{
  // remove our event from any queues
  just_remove(event);
  // get rid of all the OR events
  remove_and_free_ORs(event);
  // then fire this event
  just_fire(event);
  // now free this event
  free_event(event);
}

// --------------------------------------------------------
// new_event - create the memory for a new event and
// initialize all fields, return the new event
// --------------------------------------------------------
static struct eel_event *new_event(struct eel_group *group, 
                                   eel_event_type type, 
                                   eel_callback cb, 
                                   void *cb_arg, 
                                   int fd_or_timeout_milliseconds)
{
  // create the memory
  struct eel_event *e = (struct eel_event*)malloc(sizeof(struct eel_event));
  EEL_CHECK_NULL(new_event, e)
  if (e == NULL) { return NULL; }
  // fill in all data fields
  e->type = type;
  e->state = EEL_CREATED;
  e->callback = cb;
  e->callback_arg = cb_arg;
  e->group = group;
  // initialize the fd or timer
  if (e->type == EEL_TIMER_EVENT) {
    struct timeval now, offset;
    // compute the timer fire time    
    if (gettimeofday(&now, NULL)) {
      EEL_DEBUG_ERRORNO_MESSAGE("Failed to gettimeofday in new_event,")
      TIMERCLEAR(&e->timer);
    } else {
      MILLISECONDS_TO_TIMEVAL(offset, fd_or_timeout_milliseconds);
      TIMERADD(&now, &offset, &e->timer);
    }
  } else {
    e->fd = fd_or_timeout_milliseconds;
  }
  // initialize the OR ring
  RING_INIT(e, OR_ring);

  return e;
}

// other helpers

// --------------------------------------------------------
// free_all - clean up any events and their memory we have
// created
// --------------------------------------------------------
static void free_all()
{
  // clean up any memory we have created
  struct eel_event *e;
  struct eel_group *g;
  // free any remaining items in the eel.queue
  for (e = TAILQ_FIRST(&eel.queue); e; e = TAILQ_FIRST(&eel.queue)) {
    EEL_DEBUG_MESSAGE("Freeing extra queue entry.")
    TAILQ_REMOVE(&eel.queue, e, queue_entry);
    free_event(e);
  }
  // free any remaining items in the eel.timer_rb
  for (e = RB_MIN(event_rb_tree, &eel.timer_rb); e; e = RB_MIN(event_rb_tree, &eel.timer_rb)) {
    EEL_DEBUG_MESSAGE("Freeing extra timer entry.")
    RB_REMOVE(event_rb_tree, &eel.timer_rb, e);
    free_event(e);
  }
  // free any remaining group items in eel.group_queue
  for (g = TAILQ_FIRST(&eel.group_queue); g; g = TAILQ_FIRST(&eel.group_queue)) {
    EEL_DEBUG_MESSAGE1("Freeing extra group entry %x. eel_delete_group() not called!", (unsigned)g)
    TAILQ_REMOVE(&eel.group_queue, g, group_queue_entry);
    free_group(g);
  }
}

// --------------------------------------------------------
// process_fire - fire + cleanup all the events in 
// eel.fire_queue
// --------------------------------------------------------
static void process_fire()
{
  // fire all the events in the fire queue
  struct eel_event *e;
  for (e = TAILQ_FIRST(&eel.fire_queue); e; e = TAILQ_FIRST(&eel.fire_queue)) {
    remove_fire_and_free(e);
  }
}

// --------------------------------------------------------
// get_timeout_milliseconds - get a timeout in milliseconds
// for the given event (where its timer field is the 
// timeout time, not interval)
// --------------------------------------------------------
static int get_timeout_milliseconds(struct eel_event *e)
{
  struct timeval now, difference;
  int timeout_milliseconds = 0;
  // get the current time
  if (gettimeofday(&now, NULL)) {
    EEL_DEBUG_ERRORNO_MESSAGE("Failed to gettimeofday in get_timeout_milliseconds,")
    return 0;
  } 
  // compute the timeout value
  if (TIMERCMP(&e->timer, &now, >)) {
    TIMERSUB(&e->timer, &now, &difference);
    timeout_milliseconds = TIMEVAL_TO_MILLISECONDS(difference);
    assert(timeout_milliseconds >= 0);
  }
  return timeout_milliseconds;
}

// --------------------------------------------------------
// add_expired_timers_to_fire - add any timers that have
// expired (e.g. if several timers had the same time) to
// the fire
// --------------------------------------------------------
static void add_expired_timers_to_fire()
{
  struct timeval now;
  struct eel_event *e;
  // get the current time
  if (gettimeofday(&now, NULL)) {
    EEL_DEBUG_ERRORNO_MESSAGE("Failed to gettimeofday in add_expired_timers_to_fire,")
    return;
  } 
  // find all timers that are expired, RB_FOREACH goes min to max
  RB_FOREACH(e, event_rb_tree, &eel.timer_rb) {
    if (TIMERCMP(&e->timer, &now, >)) {
      break; // all further timers will be greater than now
    }
    add_to_fire(e);
  }
}

// --------------------------------------------------------
// event_in_progress - return 1 if the event is valid (in 
// progress), 0 if not
// --------------------------------------------------------
static inline int event_in_progress(struct eel_event *event)
{
  return (event && (event->state == EEL_WAITING || event->state == EEL_FIRE));
}

// --------------------------------------------------------
// eel_add_event_timeout - helper, add the event using
// the add_event_fn, then create a timer event and OR
// them together resulting in an event with a timeout,
// return -1 on error 0 on success
// --------------------------------------------------------
static eel_event_id eel_add_event_timeout(
                         eel_group_id gid,
                         eel_event_id(*add_event_fn)(eel_group_id gid, eel_callback cb, void *cb_arg, int fd),
                         eel_callback cb, 
                         eel_callback timeout_cb, 
                         void *cb_arg, 
                         int fd, 
                         int timeout_milliseconds)
{
  // add the event, add the timer to the same group
  struct eel_event *e1, *e2;
  e1 = add_event_fn(gid, cb, cb_arg, fd);
  if (e1 == NULL) { return NULL; }
  e2 = eel_add_timer(gid, timeout_cb, cb_arg, timeout_milliseconds);
  if (e2 == NULL) {
    just_remove(e1);
    free_event(e1);
    return NULL;
  }
  eel_make_exclusive(e1, e2);
  return e1;
}

//
// API implmentation
// 

// --------------------------------------------------------
// eel_initialize - initialize the API, return -1 on error 
// 0 on success
// --------------------------------------------------------
int eel_initialize(int exit_on_signal)
{
  EEL_DEBUG_API_CALL(eel_initialize)
  
  // set up the eel data structures
  TAILQ_INIT(&eel.queue);
  TAILQ_INIT(&eel.fire_queue);
  RB_INIT(&eel.timer_rb);
  TAILQ_INIT(&eel.group_queue);
  
  // initialize the general operators
  eel.add_to_fire = &add_to_fire;
  
  // initialize internal data fields
  eel.exit_loop = 0;
  eel.exit_on_signal = exit_on_signal;

  // get the operators
  eel.ops = get_preferred_eel_operators();
  if (eel.ops == NULL) {
    EEL_DEBUG_MESSAGE("No event mechanism available.")
    return -1;
  }
  EEL_DEBUG_MESSAGE1("Using %s as the event mechanism.", eel.ops->name)
  
  // check result
  if (eel.ops->init() == -1) {
    EEL_DEBUG_MESSAGE1("%s init failed.", eel.ops->name)
    return -1;
  }
  
  return 0;
}

// --------------------------------------------------------
// eel_uninitialize - uninitialize/cleanup the API, return 
// -1 on error 0 on success
// --------------------------------------------------------
int eel_uninitialize(void)
{
  EEL_DEBUG_API_CALL(eel_uninitialize)
  // notify the event mechanism to end
  if (eel.ops) {
    eel.ops->end();
  }
  // free any remaining items
  free_all();
  return 0;
}

// --------------------------------------------------------
// eel_dispatch_loop - the main event loop. process events
// until there are no events (return 1), error (return -1),
// or an exit was requested (return 0). In case of error,
// errno is relavent.
// --------------------------------------------------------
int eel_dispatch_loop(void)
{
  EEL_DEBUG_API_CALL(eel_dispatch)
  int return_value = 0;
  
  // loop on dispatch_once until an error, out of events
  // or exit request
  while ((return_value = eel_dispatch_once()) == 0) {
    if (eel.exit_loop) {
      eel.exit_loop = 0;
      break;
    }
  }
  
  return return_value;
}

// --------------------------------------------------------
// eel_dispatch_once - process one wait worth of events,
// return 1 if there are no events to process, -1 on error,
// or 0 on success. In case of error, errno is relavent.
// --------------------------------------------------------
int eel_dispatch_once(void)
{
  EEL_DEBUG_API_CALL(eel_dispatch_once)
  struct eel_event *lowest_timer_event = NULL;
  int timeout_milliseconds = -1;
  int wait_return;
  
  // check to make sure we have at least one event
  if (eel_no_events()) {
    return 1;
  }
  // find the lowest timer
  lowest_timer_event = RB_MIN(event_rb_tree, &eel.timer_rb);
  if (lowest_timer_event) {
    timeout_milliseconds = get_timeout_milliseconds(lowest_timer_event);
  }
  
  // wait on the events and lowest timer
  wait_return = eel.ops->wait(&eel, timeout_milliseconds);
  // see what happened
  if (wait_return == -1) {
    if (errno == EINTR) {
      // we were interupted by a signal, continue unless we have been instructed to stop
      if (eel.exit_on_signal) {
        EEL_DEBUG_MESSAGE("Received signal, exiting loop.")
        return -1;
      }
      EEL_DEBUG_MESSAGE("Received signal, continuing.")
    } else {
      // there was an actual error
      EEL_DEBUG_ERRORNO_MESSAGE("wait() returned -1,");
      return -1;
    }
  } else if (wait_return == 0 && lowest_timer_event) {
    // the timer is what fired, make sure it gets added
    add_to_fire(lowest_timer_event);
  } 
  // add any timers that have expired to the fire queue
  add_expired_timers_to_fire();
  // fire all the events in the fire queue
  process_fire();
  // success
  return 0;
}

// --------------------------------------------------------
// eel_no_events - return non-0 if there are events, 0
// if none are present
// --------------------------------------------------------
int eel_no_events(void)
{
  //EEL_DEBUG_API_CALL(eel_no_events) - do not log this API call
  return ( TAILQ_EMPTY(&eel.queue) && RB_EMPTY(&eel.timer_rb) );
}

// --------------------------------------------------------
// eel_exit_loop - request eel_dispatch_loop to exit on 
// next pass
// --------------------------------------------------------
void eel_exit_loop(void)
{
  EEL_DEBUG_API_CALL(eel_exit_loop)
  eel.exit_loop = 1;
}

// --------------------------------------------------------
// eel_add_timer - add a timer event, fill in handle if not
// NULL, return -1 on error 0 on success
// --------------------------------------------------------
eel_event_id eel_add_timer(eel_group_id gid, eel_callback cb, void *cb_arg, int timeout_milliseconds)
{
  EEL_DEBUG_API_CALL(eel_add_timer)
  struct eel_event *e = new_event(gid, EEL_TIMER_EVENT, cb, cb_arg, timeout_milliseconds);
  add_to_queue(e);
  add_to_group(gid, e); 
  // do not add timers to event mechanism
  return e;
}

// --------------------------------------------------------
// eel_add_read - add a read event, fill in handle if not
// NULL, return -1 on error 0 on success
// --------------------------------------------------------
eel_event_id eel_add_read(eel_group_id gid, eel_callback cb, void *cb_arg, int fd)
{
  EEL_DEBUG_API_CALL(eel_add_read)
  struct eel_event *e = new_event(gid, EEL_READ_EVENT, cb, cb_arg, fd);
  add_to_queue(e);
  add_to_group(gid, e);
  eel.ops->add(&eel, e); //*** check return value here
  return e;
}

// --------------------------------------------------------
// eel_add_write - add a write event, fill in handle if not
// NULL, return -1 on error 0 on success
// --------------------------------------------------------
eel_event_id eel_add_write(eel_group_id gid, eel_callback cb, void *cb_arg, int fd)
{
  EEL_DEBUG_API_CALL(eel_add_write)
  struct eel_event *e = new_event(gid, EEL_WRITE_EVENT, cb, cb_arg, fd);
  add_to_queue(e);
  add_to_group(gid, e);
  eel.ops->add(&eel, e); //*** check return value here
  return e;
}

// --------------------------------------------------------
// eel_add_error - add a error event, fill in handle if not
// NULL, return -1 on error 0 on success
// --------------------------------------------------------
eel_event_id eel_add_error(eel_group_id gid, eel_callback cb, void *cb_arg, int fd)
{
  EEL_DEBUG_API_CALL(eel_add_error)
  struct eel_event *e = new_event(gid, EEL_ERROR_EVENT, cb, cb_arg, fd);
  add_to_queue(e);
  add_to_group(gid, e);
  eel.ops->add(&eel, e); //*** check return value here
  return e;
}

// Compound functions

// --------------------------------------------------------
// eel_add_read_timeout - add a read event, plus an ORed
// timer event, fill in handle if not NULL, return -1 on 
// error 0 on success. 
// --------------------------------------------------------
eel_event_id eel_add_read_timeout(eel_group_id gid, 
                         eel_callback cb, 
                         eel_callback timeout_cb, 
                         void *cb_arg, 
                         int fd, 
                         int timeout_milliseconds)
{
  EEL_DEBUG_API_CALL(eel_add_read_timeout)
  return eel_add_event_timeout(gid, &eel_add_read, cb, timeout_cb, cb_arg, fd, timeout_milliseconds);
}

// --------------------------------------------------------
// eel_add_write_timeout - add a write event, plus an ORed
// timer event, fill in handle if not NULL, return -1 on 
// error 0 on success. 
// --------------------------------------------------------
eel_event_id eel_add_write_timeout(eel_group_id gid, 
                         eel_callback cb, 
                         eel_callback timeout_cb, 
                         void *cb_arg, 
                         int fd, 
                         int timeout_milliseconds)
{
  EEL_DEBUG_API_CALL(eel_add_write_timeout)
  return eel_add_event_timeout(gid, &eel_add_write, cb, timeout_cb, cb_arg, fd, timeout_milliseconds);
}

// --------------------------------------------------------
// eel_add_error_timeout - add a error event, plus an ORed
// timer event, fill in handle if not NULL, return -1 on 
// error 0 on success. 
// --------------------------------------------------------
eel_event_id eel_add_error_timeout(eel_group_id gid, 
                         eel_callback cb, 
                         eel_callback timeout_cb, 
                         void *cb_arg, 
                         int fd, 
                         int timeout_milliseconds)
{
  EEL_DEBUG_API_CALL(eel_add_error_timeout)
  return eel_add_event_timeout(gid, &eel_add_error, cb, timeout_cb, cb_arg, fd, timeout_milliseconds);
}

// Event operations

// --------------------------------------------------------
// eel_make_exclusive - 
// --------------------------------------------------------
// Events must be in the same group to be exclusive, as well events must be valid
int eel_make_exclusive(eel_event_id eid_in_out, eel_event_id eid_in)
{
  if (eid_in_out == NULL || eid_in == NULL || 
      eid_in_out->group != eid_in->group) {
    return -1;
  }
  // create/merge the OR rings of the two events
  RING_INSERT(eel_event, OR_ring, eid_in_out, eid_in);
  return 0;
}

// Group Operations

// --------------------------------------------------------
// eel_new_group_id - 
// --------------------------------------------------------
eel_group_id eel_new_group_id(void)
{
  EEL_DEBUG_API_CALL(eel_new_group_id)
  struct eel_group *group = (struct eel_group*)calloc(1, sizeof(struct eel_group));
  EEL_CHECK_NULL(eel_new_group_id, group)
  if (group == NULL) { return NULL; }
  TAILQ_INSERT_TAIL(&eel.group_queue, group, group_queue_entry);
  TAILQ_INIT(&group->event_queue);
  return group;
}

// --------------------------------------------------------
// eel_delete_group_id - 
// --------------------------------------------------------
void eel_delete_group_id(eel_group_id gid)
{
  EEL_DEBUG_API_CALL(eel_delete_group_id)
  eel_remove_events(gid);
  TAILQ_REMOVE(&eel.group_queue, gid, group_queue_entry);
  free_group(gid);
}

// --------------------------------------------------------
// eel_has_events - return 1 if the event is valid (in 
// progress), 0 if not
// --------------------------------------------------------
int eel_has_events(eel_group_id gid)
{
  struct eel_event *e;
  int return_value = 0;
  EEL_DEBUG_API_CALL(eel_has_events)
  if (gid == NULL) { return 0; }
  TAILQ_FOREACH(e, &gid->event_queue, group_queue_entry) {
    return_value |= event_in_progress(e);
  }
  return return_value;
}

// --------------------------------------------------------
// eel_remove_events - remove all the events specified by 
// gid, return -1 on error 0 on success
// --------------------------------------------------------
int eel_remove_events(eel_group_id gid)
{
  struct eel_event *e;
  EEL_DEBUG_API_CALL(eel_remove_events)
  if (gid == NULL) { return -1; }
  for (e = TAILQ_FIRST(&gid->event_queue); e; e = TAILQ_FIRST(&gid->event_queue)) {
    just_remove(e);
    remove_and_free_ORs(e);
    free_event(e);
  }
  return 0;
}
