//
// C Implementation: eel_select
//
// Description: 
//
//
// Author: Ryan Cunningham <rcunning@cs.ucla.edu>, (C) 2004
//
// Copyright: Ryan Cunningham
//
//

// a mess of headers here, likely can remove some
#include <sys/types.h>
#include <tree.h>
#include <queue.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <err.h>
#include <assert.h>

#include <sys/select.h>

#include "eel.h"
#include "eel_util.h"
#include "eel_select.h"

//--- check EEL_NO_SELECT macro

// fd set size helper macro
#ifndef howmany
#define        howmany(x, y)   (((x)+((y)-1))/(y))
#endif

// prototypes for eel_operators implementation
int   select_init       ();
void  select_end        ();
int   select_add        (struct eel_lib_struct *eel, struct eel_event *e);
int   select_remove     (struct eel_lib_struct *eel, struct eel_event *e);
int   select_wait       (struct eel_lib_struct *eel, int timeout_milliseconds);

// eel operator instance
struct eel_operators select_operators = {
  "select",
  &select_init,
  &select_end,
  &select_add,
  &select_remove,
  &select_wait
};

// --------------------------------------------------------
// get_select_eel_operators - return our operators
// --------------------------------------------------------
struct eel_operators *get_select_eel_operators()
{
  if (getenv("EEL_NO_SELECT")) {
    return NULL;
  }
  return &select_operators;
}

// structure used to save data between eel operator calls
struct select_data_struct {
  fd_set  *read_set;    // the read events fd_set
  fd_set  *write_set;   // the write events fd_set
  fd_set  *error_set;   // the error events fd_set
  int     fd_set_size;  // the size of each of the above sets
  int     max_fd;       // the maximum fd value
} select_data;

//
// Implementation
//

// --------------------------------------------------------
// select_init - initialize
// --------------------------------------------------------
int select_init()
{
  // clear out our structure
  memset(&select_data, 0, sizeof(select_data));
  return 0;
}

// --------------------------------------------------------
// select_end - clean up
// --------------------------------------------------------
void select_end()
{
  // free the memory created for the sets
  SAFE_FREE(select_data.read_set);
  SAFE_FREE(select_data.write_set);
  SAFE_FREE(select_data.error_set);
}

// --------------------------------------------------------
// select_add - an event was added, keep track of highest
// fd
// --------------------------------------------------------
int select_add(struct eel_lib_struct *eel, struct eel_event *e)
{
  if (e->fd > select_data.max_fd) {
    select_data.max_fd = e->fd;
  }
  return 0;
}

// --------------------------------------------------------
// select_remove - an event was removed, do nothing
// --------------------------------------------------------
int select_remove(struct eel_lib_struct *eel, struct eel_event *e)
{
  // do nothing -- don't bother reducing the set size when fds are removed
  return 0;
}

// --------------------------------------------------------
// select_update_size - helper function to grow the fd_sets
// --------------------------------------------------------
static int select_update_size()
{
  int new_fd_set_size;
  // get the new size
  new_fd_set_size = howmany(select_data.max_fd + 1, NFDBITS) * sizeof(fd_mask);
  if (new_fd_set_size > select_data.fd_set_size)
  {
    // if we need more space, realloc to the new size
    select_data.read_set  = (fd_set*)realloc(select_data.read_set, new_fd_set_size);
    select_data.write_set = (fd_set*)realloc(select_data.write_set, new_fd_set_size);
    select_data.error_set = (fd_set*)realloc(select_data.error_set, new_fd_set_size);
    if (select_data.read_set == NULL ||
        select_data.write_set == NULL ||
        select_data.error_set == NULL) {
        EEL_DEBUG_MESSAGE("realloc() returned NULL in select_update_size.")
        return -1;
    }
    // save the new size
    select_data.fd_set_size = new_fd_set_size;
  }
  return 0;
}

// --------------------------------------------------------
// select_wait - implement wait functionality
// --------------------------------------------------------
int select_wait(struct eel_lib_struct *eel, int timeout_milliseconds)
{
  int select_result = 0;
  int new_max_fd = 0;
  struct timeval timeout;
  struct timeval *ptimeout = NULL;
  struct eel_event *e;
  
  // update the stucture size
  if (select_update_size() == -1) {
    return -1;
  }
  // clear out the sets
  memset(select_data.read_set, 0, select_data.fd_set_size);
  memset(select_data.write_set, 0, select_data.fd_set_size);
  memset(select_data.error_set, 0, select_data.fd_set_size);
  // fill the sets with the current fds, note: O(n)
  TAILQ_FOREACH(e, &eel->queue, queue_entry) {
    switch (e->type) {
    case EEL_READ_EVENT:
      FD_SET(e->fd, select_data.read_set);
      break;
    case EEL_WRITE_EVENT:
      FD_SET(e->fd, select_data.write_set);
      break;
    case EEL_ERROR_EVENT:
      FD_SET(e->fd, select_data.error_set);
      break;
    case EEL_TIMER_EVENT:
      assert(1);
      break;
    };
  }
  
  // set up the timeout if necessary
  if (timeout_milliseconds != -1) {
    MILLISECONDS_TO_TIMEVAL(timeout, timeout_milliseconds);
    ptimeout = &timeout;
  }
  
  // do the select
  select_result = select( select_data.max_fd + 1, 
                          select_data.read_set, 
                          select_data.write_set, 
                          select_data.error_set, 
                          ptimeout);
  
  // add events to fire queue, note: O(n)
  if (select_result > 0) {
    TAILQ_FOREACH(e, &eel->queue, queue_entry) {
      if ( ((e->type == EEL_READ_EVENT) && (FD_ISSET(e->fd, select_data.read_set))) || 
           ((e->type == EEL_WRITE_EVENT) && (FD_ISSET(e->fd, select_data.write_set))) || 
           ((e->type == EEL_ERROR_EVENT) && (FD_ISSET(e->fd, select_data.error_set))) )
      {
        eel->add_to_fire(e);
      } else {
        if (e->fd > new_max_fd) {
          new_max_fd = e->fd;
        }
      }
    }
    // update the max_fd to be the highest of the events that did not fire
    select_data.max_fd = new_max_fd;
  }
  
  return select_result;
}
